#-*-coding:utf-8-*-

__all__ = ['TaskBase']

import os
import re
from collections  import defaultdict
import json
import arrow
from inflector    import Inflector
from etl_utils    import cached_property
from dateutil     import tz

from ...luigi_ext        import luigi
from ..other.root_task   import RootTask
from ...utils            import DateUtils, ExtUtils, IOUtils
from ...parameter        import ArrowParameter
from ...manager          import luiti_config


class TaskBase(luigi.Task, ExtUtils.ExtendClass):
    """ 继承的子类在类名后 必须加 **时间类型**, 如 Day, Week, ... """

    run_mode           = ["local", "mr_distribute", "mr_local"][0]

    date_value         = ArrowParameter() # **统一** 时间类型, 防止同时跑多个任务
    orig_date_value    = None

    # will overwritten by @decorator
    # 不能以 **两个 __ 开头**, 否则会被 Python 当作隐私变量而无法继承。TODO 隐私变量 可能是错的。
    _persist_files    = []
    _ref_tasks        = []

    root_dir           = NotImplemented

    # Default one, always return True
    def requires(self):
        return RootTask()

    def run     (self):
        raise NotImplementedError

    def __init__(self, *args, **kwargs):
        super(TaskBase, self).__init__(*args, **kwargs)

        # 在跨期的时候用于判断 该周应该是该周的哪些天。
        # 比如这学期开学是 2015-02-17(星期二) 开学, 那么这周的数据只有 0217-0222。
        # 而在寒假里(即run 2015-02-16(星期天) 的 task 时，那么该周的天只有 0216 一天。
        self.orig_date_value = ArrowParameter.get(self.date_value).replace(tzinfo=tz.tzlocal())

        self.reset_date() # reset date to at the beginning of current date type here
        self.data_file      # force load it now, or `output` still load it.
        self.package_name   # force load it now, use to serialize


    @cached_property
    def data_dir(self):
        assert self.root_dir
        return os.path.join(self.root_dir, self.date_str)

    @cached_property
    def data_file(self):
        return os.path.join(self.data_dir, self.data_name + ".json")

    @cached_property
    def data_name(self):
        return Inflector().underscore(self.__class__.__name__)

    def output(self):
        return IOUtils.local_target(self.data_file)

    def errput(self):
        return IOUtils.local_target(self.data_file + ".err")

    @cached_property
    def date_str(self):
        return self.date_value.strftime("%Y-%m-%d")

    @cached_property
    def date_type(self):
        return luiti_config.get_date_type(self.__class__.__name__)

    @cached_property
    def date_value_by_type_in_last(self):
        return DateUtils.date_value_by_type_in_last(self.date_value, self.date_type)

    @cached_property
    def date_value_by_type_in_begin(self):
        return ArrowParameter.get(self.date_value).floor(self.date_type)

    @cached_property
    def date_value_by_type_in_end(self):
        return ArrowParameter.get(self.date_value).ceil(self.date_type)

    @cached_property
    def pre_task_by_self(self):
        """ 如果跨了两个周期就没有上次数据文件了 """
        return RootTask() if self.is_reach_the_edge else self.__class__(self.date_value_by_type_in_last)

    @cached_property
    def is_reach_the_edge(self):
        return False # default. e.g. add semester

    def reset_date(self):
        # **强制** 写为统一时间格式(arrow格式)，这样luigi就不会同时跑两个任务了。
        self.date_value = ArrowParameter.get(self.date_value)

        orig_date       = self.date_value
        if self.date_type != 'range':
            new_date = orig_date.floor(self.date_type)
            if orig_date != new_date:
                print "[reset date by %s] from %s to %s" % (self.date_type, orig_date, new_date)
                self.date_value = new_date


    @classmethod
    def instances_by_date_range(cls, first_date, last_date):
        """ 返回属于某周期里的所有当前任务实例列表 """
        assert isinstance(first_date, arrow.Arrow)
        assert isinstance(last_date,  arrow.Arrow)

        if "Range" in cls.__name__:
            return list(set([cls(first_date), cls(last_date)])) # return head and tail directly
        else:
            dates = arrow.Arrow.range(luiti_config.get_date_type(cls.__name__), first_date, last_date)
            return [cls(date1.datetime) for date1 in dates]

    @cached_property
    def task_class(self):
        return self.__class__

    @cached_property
    def package_name(self):
        module_name  = self.task_class.__module__
        package_name = module_name.split(".")[0]
        return package_name



#-*-coding:utf-8-*-

import os
import commands

class CommandUtils:

    @staticmethod
    def execute(command_str, dry=False, verbose=True):
        if verbose:   print "[command]", command_str
        if dry:       return False

        #return commands.getstatusoutput(command_str)
        return os.system(command_str) # print logs in realtime.



#-*-coding:utf-8-*-

__all__ = ['HadoopExt']

import luigi.hadoop
from .utils import ExtUtils, TargetUtils
from .parameter import ArrowParameter


class LuitiHadoopJobRunner(luigi.hadoop.HadoopJobRunner):
    """ overwrite DefaultHadoopJobRunner.class """

    # params are copied from HadoopJobRunner
    def __init__(self, libjars=None, output_format=None):
        config = luigi.hadoop.configuration.get_config()
        opts = {
                "streaming_jar"         : config.get('hadoop', 'streaming-jar'),
                "output_format"         : output_format,
                "libjars"               : libjars,
            }
        super(LuitiHadoopJobRunner, self).__init__(**opts)


class HadoopExt(luigi.hadoop.JobTask, ExtUtils.ExtendClass):

    run_mode           = "mr_distribute"

    n_reduce_tasks    = 1 # 体现在 输出的part-00000数量为reduce数量

    physicalmemory    = 2
    reduce_memory_GB  = 2
    (  map_memory_GB) = 0.6
    iosort_memory_GB  = 0.5

    output_format     = [
                            #"org.apache.hadoop.mapreduce.lib.output.TextOutputFormat",         # 单路输出。这个版本有问题。
                            "org.apache.hadoop.mapred.TextOutputFormat",                        # 单路输出
                            "org.apache.hadoop.mapred.lib.MultipleTextOutputFormat",            # 多路输出
                        ][0]                                                                    # 默认是 单路输出
    libjars           = []


    def __init__(self, *args, **kwargs):
        """ 参考 TaskBase, 确保在 继承时还可以有TaskBase的覆写日期功能。 """
        super(HadoopExt, self).__init__(*args, **kwargs)
        self.orig_date_value = self.orig_date_value or ArrowParameter.get(self.date_value)
        self.reset_date()

    def output(self):
        return luigi.HDFS(self.data_file)

    # overwrite
    def job_runner(self):
        """ will be wraped in `run` function. """
        return LuitiHadoopJobRunner(output_format=self.output_format, libjars=self.libjars)

    def output(self):
        return TargetUtils.hdfs(self.data_file)

    def jobconfs_opts(self):
        return [
                "mapreduce.framework.name=yarn",
                'mapred.reduce.tasks=%s' % self.n_reduce_tasks,
               ]

    def jobconfs(self):
        jcs = super(luigi.hadoop.JobTask, self).jobconfs()
        for conf_opt_1 in self.jobconfs_opts():
            jcs.append(conf_opt_1)
        return jcs



    # TestCase related attrs
    def mrtest_input(self):
        raise NotImplementedError

    def mrtest_output(self):
        raise NotImplementedError

    def mrtest_attrs(self):
        return dict()



#-*-coding:utf-8-*-

import os
import sys
import string
import traceback
import importlib
from inflector import Inflector

from .config            import luiti_config as lc
from .active_packages   import active_packages
from .package_map       import PackageMap



class Loader(object):

    @staticmethod
    @active_packages
    def load_all_tasks():
        result     = {"success": list(), "failure": list()}

        for task_clsname_1 in sorted(PackageMap.task_clsname_to_package.keys()):
            is_success = False
            task_cls   = None
            err        = None

            try:
                task_cls    = Loader.load_a_task_by_name(task_clsname_1)
                is_success  = True
            except Exception as e:
                err = list(sys.exc_info())
                err[2] = "".join(traceback.format_tb(err[2]))
                err = str(err[0]) + ": " + str(err[1]) + "\n" + err[2]

            if is_success:
                result['success'].append({"task_cls": task_cls})
            else:
                result['failure'].append({"err": err,           "task_clsname": task_clsname_1})

        return result

    @staticmethod
    @active_packages
    def load_a_task_by_name(s1):
        task_clsname_1  = Inflector().classify(s1)    # force convert
        task_filename_1 = Inflector().underscore(s1)  # force convert

        assert task_clsname_1 in PackageMap.task_clsname_to_package, u"""
        "%s" cannt be found. Auto converted class name is "%s", file name is "luiti_tasks/%s.py", please check it carefully.
        """ % (s1, task_clsname_1, task_filename_1)

        package_path = PackageMap.task_clsname_to_package[task_clsname_1].__name__ + \
                       ".luiti_tasks." + task_filename_1
        task_lib     = lc.import2(package_path)
        return getattr(task_lib, task_clsname_1)



#-*-coding:utf-8-*-

__all__ = ['TaskRangeHadoop', 'luigi', 'MRUtils', 'json', 're']

from .task_range import TaskRange, luigi

from ...utils import MRUtils
import json
import re

class TaskRangeHadoop(luigi.hadoop.HadoopExt, TaskRange):

    pass



#-*-coding:utf-8-*-

import os, sys
from inflector    import Inflector
from etl_utils import singleton, cached_property
import arrow

@singleton()
class LuitiConfigClass(object):
    """ Make sure init variables only once. """
    DateTypes          = ["range", "week"] + arrow.Arrow._ATTRS # ['year', 'month', ...]

    curr_project_name = None
    curr_project_dir  = None

    linked_luigi      = None

    @cached_property
    def attached_package_names(self):
        return set(['luiti'])

    @cached_property
    def luiti_tasks_packages(self):
        return set([])

    @staticmethod
    def import2(a_package):
        return __import__(a_package, None, None, 'non_empty')

    @staticmethod
    def get_date_type(name1):
        """ Inherit class must be in TaskBase{Day,Week,Month,Range} style.  """
        assert isinstance(name1, (str, unicode))
        str1 = Inflector().underscore(name1).split("_")[-1].lower()
        assert str1 in luiti_config.DateTypes
        return str1

    @staticmethod
    def get_time_task(name1):
        """ return e.g. TaskDay """
        type2 = luiti_config.get_date_type(name1)
        return "Task" + Inflector().camelize(type2)

    @staticmethod
    def link_packages():
        """
        called by `active_packages`
        """
        lc = luiti_config
        is_in_luigi_distributed = False

        # 1. unmornal task class
        if lc.curr_project_name == "__main__": return False

        # 2. setup current project as root
        if lc.curr_project_dir is None: lc.curr_project_dir = os.getcwd() # auto from current class
        lc.fix_project_dir()

        def exists(filename1):
            return os.path.exists(os.path.join(lc.curr_project_dir, filename1))

        # These files are created by luigi.
        if exists("job-instance.pickle") and exists("job.jar") and exists("packages.tar") and exists("luigi"):
            is_in_luigi_distributed = True

        is_a_luiti_project = exists("luiti_tasks") # compact with no-luiti project

        if lc.curr_project_name is None:
            # Check current work dir is under a valid a luiti_tasks project
            #if not is_in_luigi_distributed:
            #    raise ValueError(u"""[error] current work dir [%s] has no luiti_tasks dir! It has these files/dirs %s""" % (lc.curr_project_dir,
            #               os.listdir(lc.curr_project_dir),))

            if is_in_luigi_distributed:
                for item1 in os.listdir(lc.curr_project_dir):
                    if exists(item1 + "/__init__.py") and exists(item1 + "/luiti_tasks"): # is a valid python package
                        lc.luiti_tasks_packages.add(lc.import2(item1))
            else:
                curr_project_name     = os.path.basename(lc.curr_project_dir) # "project_A"
                lc.curr_project_name  = curr_project_name

                curr_project_syspath  = os.path.dirname(lc.curr_project_dir)  # project_A/
                if curr_project_syspath not in sys.path: sys.path.insert(0, curr_project_syspath)

                lc.luiti_tasks_packages.add(lc.import2(lc.curr_project_name))

                # 3. ensure other luiti tasks packages can be loaded.
                if is_a_luiti_project:
                    lc.import2(lc.curr_project_name + ".luiti_tasks.__init_luiti")

    def fix_project_dir(self):
        """ Fix project_A/project_A/luiti_tasks dir """
        _try_dir   = os.path.join(luiti_config.curr_project_dir, os.path.basename(luiti_config.curr_project_dir))
        if os.path.exists(_try_dir): # cause of the same name
            luiti_config.curr_project_dir = _try_dir




luiti_config = LuitiConfigClass()



#-*-coding:utf-8-*-

__all__ = ['env']

import os, sys
from etl_utils import singleton, cached_property

@singleton()
class EnvClass(object):
    """ Setup luigi global env. """

    is_setuped = False

    sys_path = []

    def setup(self):
        if self.is_setuped is True:
            return False
        else:
            self.setup_sys_path
            self.is_setuped = True
            return True

    @cached_property
    def setup_sys_path(self):
        assert len(env.sys_path), "[error] Must setup Python sys path at least one path! e.g. `luiti.env.sys_path = ['some_python_library_dir', ...]` "

        # init only once
        if "path_old_by_luiti" in dir(sys):
            raise ValueError("[error] Already setup!")
        else:
            sys.path_old_by_luiti = sys.path
            sys.path     = []
            for path1 in (env.sys_path + sys.path_old_by_luiti):
                if path1 not in sys.path:
                    sys.path.append(path1)

        return sys.path

env = EnvClass()
# TODO 也许挪到 .manager



#-*-coding:utf-8-*-

import json
import luigi
from etl_utils import singleton, cached_property


class TargetUtils:

    @staticmethod
    def line_read(hdfs1):
        with hdfs1.open('r') as data1:
            for line1 in data1:
                yield line1.decode("UTF-8").strip()

    @staticmethod
    def json_read(hdfs1):
        for line1 in TargetUtils.line_read(hdfs1):
            yield json.loads(line1) # as item1

    @staticmethod
    def hdfs(data_file1):
        # [兼容] 可以判断出 data_file1 是否包含 part-00000 的目录。

        # 兼容 snakebite 对 不存在目录的 test 有bug，或者是因为从hadoop用户切换到primary_user导致。
        f1 = luigi.hdfs.HdfsTarget(data_file1)
        if f1.exists():
            is_curr_dir = len(list(f1.fs.listdir(data_file1))) > 1 # isdir 在 luigi/hdfs.py 没有实现哦

            if is_curr_dir:
                if luigi.hdfs.HdfsTarget(data_file1 + "/part-00000").exists() and \
                    luigi.hdfs.HdfsTarget(data_file1 + "/_SUCCESS").exists():
                    return luigi.hdfs.HdfsTarget(data_file1, format=luigi.hdfs.PlainDir)

        return luigi.hdfs.HdfsTarget(data_file1)

    @staticmethod
    def mr_read(hdfs1):
        from .mr_utils import MRUtils
        for line1 in TargetUtils.line_read(hdfs1):
            k_1, v_1 = MRUtils.split_mr_kv(line1)
            yield k_1, v_1

    @staticmethod
    def isdir(path1): return HdfsClient.client.get_bite().test(path1, directory=True)

    @staticmethod
    def exists(path1): return HdfsClient.client.exists(path1)



@singleton()
class HdfsClientClass(object):
    # TODO use delegate
    @cached_property
    def client(self):
        import luigi.hdfs
        return luigi.hdfs.client
HdfsClient = HdfsClientClass()
TargetUtils.HdfsClient = HdfsClient



#-*-coding:utf-8-*-

from .io_utils           import IOUtils
from .target_utils       import TargetUtils
from .mr_utils           import MRUtils
from .math_utils         import MathUtils
from .hdfs_utils         import HDFSUtils
from .command_utils      import CommandUtils
from .date_utils         import DateUtils
from .compress_utils     import CompressUtils
from .ext_utils          import ExtUtils



#-*-coding:utf-8-*-

import json
import luigi, luigi.hdfs
from luigi import LocalTarget
from etl_utils import JsonUtils

class IOUtils:

    SQL_RANGE_LIMIT = 1000

    @staticmethod
    def write_json_to_output(json1, output1):
        with output1.open('w') as output_hdfs:
            if isinstance(json1, (list,set,)):
                method = lambda item1: json.dumps(list(item1)) # 兼容 JsonUtils.unicode_dump 不支持list
            else:
                method = lambda item1: JsonUtils.unicode_dump(item1).encode("UTF-8")
            output_hdfs.write(method(json1) + "\n")

    @staticmethod
    def read_json_from_output(output1):
        # only one line
        item1 = None
        with output1.open('r') as output_hdfs:
            for line1 in output_hdfs:
                item1 = json.loads(line1)
        return item1

    @staticmethod
    def remove_files(*files): # 兼容 写入中途失败
        for file1 in files:
            if luigi.hdfs.exists(file1):
                luigi.hdfs.remove(file1)

    @staticmethod
    def local_target(path1): return LocalTarget(path1)



#-*-coding:utf-8-*-

__all__ = [
            'luigi', 'luiti_setup', 'config',

            'TaskBase',
            'TaskHour',
            'TaskDay',
            'TaskDayHadoop',
            'TaskWeek',
            'TaskWeekHadoop',
            'TaskMonth',
            'TaskRange',
            'TaskRangeHadoop',

            'RootTask',
            'StaticFile',
            'MongoImportTask',
            'HadoopExt',

            'manager',

            'IOUtils', 'DateUtils', 'TargetUtils', 'HDFSUtils',
            'MRUtils', 'MathUtils', 'CommandUtils',
            'CompressUtils',

            'ArrowParameter',

            'os', 're', 'defaultdict', 'json', 'cached_property',
            'arrow',

            'MrTestCase',
          ]

import os
import sys
import re
from collections import defaultdict
import json
import arrow
from etl_utils import cached_property

from .luigi_ext            import luigi
from .hadoop_ext           import HadoopExt
from .env                  import env

from .task_templates.time.task_base            import TaskBase
from .task_templates.time.task_hour            import TaskHour
from .task_templates.time.task_day             import TaskDay
from .task_templates.time.task_week            import TaskWeek
from .task_templates.time.task_month           import TaskMonth
from .task_templates.time.task_range           import TaskRange

from .task_templates.time.task_day_hadoop      import TaskDayHadoop
from .task_templates.time.task_week_hadoop     import TaskWeekHadoop
from .task_templates.time.task_range_hadoop    import TaskRangeHadoop

from .task_templates.other.root_task           import RootTask
from .task_templates.other.static_file         import StaticFile
from .task_templates.other.mongo_import_task   import MongoImportTask

from .                     import manager
from .utils                import IOUtils, DateUtils, TargetUtils, HDFSUtils, \
                                  MRUtils, MathUtils, CommandUtils, \
                                  CompressUtils

from .parameter            import ArrowParameter

from .mr_test_case         import MrTestCase



# TODO 整理
def luiti_setup(opts=None):
    # 1. default value
    assert 'sys_path' in opts, "opts['sys_path'] must be configed! Or there will is none sys.path existed in distributed Hadoop servers."
    default_opts = {
            'sys_path'                : [],
            'LUIGI_CONFIG_PATH'       : '/etc/luigi/client.cfg',
        }
    opts = dict(default_opts.items() + (opts or dict()).items())

    # 2. config
    env.sys_path                      = opts['sys_path']
    os.environ['LUIGI_CONFIG_PATH']   = opts['LUIGI_CONFIG_PATH']

    # 3. setup env
    env.setup()

config = manager.luiti_config



#-*-coding:utf-8-*-

from .loader              import Loader
from .table               import Table
from .dep                 import Dep
from .files               import Files

from .config              import luiti_config
from .package_map         import PackageMap
from .active_packages     import active_packages

from .sys_argv            import SysArgv

from .generate_from_templates import GenerateFromTemplates



#################
### API list  ###
#################

find_dep_on_tasks                                 = Dep.find_dep_on_tasks
get_all_date_file_to_task_instances               = Files.get_all_date_file_to_task_instances
soft_delete_files                                 = Files.soft_delete_files
load_all_tasks                                    = Loader.load_all_tasks
load_a_task_by_name                               = Loader.load_a_task_by_name
import2                                           = luiti_config.import2
print_all_tasks                                   = Table.print_all_tasks
print_files_by_task_cls_and_date_range            = Table.print_files_by_task_cls_and_date_range
new_a_project                                     = GenerateFromTemplates.new_a_project
generate_a_task                                   = GenerateFromTemplates.generate_a_task


from etl_utils import singleton, cached_property
@singleton()
class LazyData(object):

    @cached_property
    def all_task_classes(self):
        return [i1['task_cls'] for i1 in ld.result['success']]

    @cached_property
    def result(self):
        return load_all_tasks()

ld = LazyData()



#-*-coding:utf-8-*-

__all__ = ['luigi']

import luigi
from luigi import Event
import luigi.hadoop
luigi.hadoop = luigi.hadoop
import luigi.hdfs
luigi.hdfs = luigi.hdfs

from .hadoop_ext import HadoopExt
luigi.hadoop.HadoopExt = HadoopExt # write back
# NOTE 对 luigi.hadoop 兼容 "track the job: "


import os, sys
import arrow
from .parameter import ArrowParameter
import importlib
from inflector import Inflector
from etl_utils import cached_property, singleton
from collections import defaultdict

from .utils import IOUtils, TargetUtils
from . import manager
luigi.HDFS = TargetUtils.hdfs # 本来就是需要读取全局配置，所以索性就绑定在 luigi 命名空间了吧。

def persist_files(*files): # 装饰器
    """ 多个data_file 可以用 DSL 描述，然后和 event_handler(Event.FAILURE) 绑定在一起 """
    def func(cls):
        # 1. 设置 持久化文件属性
        def wrap(file1): # 这样才可以保存 file1 变量，而不至于被覆写。
            def _file(self):
                return os.path.join(self.data_dir, file1 + ".json")
            return _file

        cls.__persist_files = files
        for file1 in cls.__persist_files:
            setattr(cls, file1, property(wrap(file1))) # @decorator

        # 2. 绑定 失败时删除这些文件
        def clean_tmp(task, exception):
            for file1 in files: IOUtils.remove_files(getattr(task, file1))
            # IOUtils.remove_files(task.data_file) # NOTE 好像 Hadoop 会自动处理失败任务的输出文件的，否则就会导致其在N次重试一直在running。
        cls.event_handler(Event.FAILURE)(clean_tmp)

        return cls

    return func
luigi.persist_files = persist_files



def ref_tasks(*tasks): # 装饰器
    """
    自动把依赖 Task 链接起来，通过属性访问。

    Example:

    ```python
    @ref_tasks("TaskB", "TaskC")
    class TaskA(TaskWeekBase):
        pass

    TaskA().TaskB == TaskB
    TaskA().TaskC == TaskC
    ```
    """
    def wrap_cls(ref_task_name):
        def _func(self):
            v1 = self.__dict__.get(ref_task_name, None)
            if v1 is None:
                v1 = manager.load_a_task_by_name(ref_task_name)
                self.__dict__[ref_task_name] = v1
            return v1
        return _func

    def wrap_instance(ref_task_name, task_name):
        def _func(self):
            v1 = self.__dict__.get(task_name, None)
            if v1 is None:
                v1 = getattr(self, ref_task_name)(self.date_value)
                self.__dict__[task_name] = v1
            return v1
        return _func

    def __getstate__(self):
        """ Fix luiti_tasks module namespace conflicts. """
        for ref_task1 in self._ref_tasks:
            cname = ref_task1           # class    name
            iname = ref_task1 + "_task" # instance name

            # delete instance property is enough.
            #if hasattr(self.__class__, cname):  delattr(self.__class__, cname)
            #if hasattr(self.__class__, iname):  delattr(self.__class__, iname)

            if cname in self.__dict__:          del self.__dict__[cname]
            if iname in self.__dict__:          del self.__dict__[iname]
        return self.__dict__

    def __setstate__(self, d1):
        # 1. default
        self.__dict__.update(d1)
        # 2. plug other package in `.__init_luiti`
        luigi.luiti_config.curr_project_name = self.package_name
        luigi.luiti_config.link_packages()

# cached_property 捕获不了 ref_task_name 变量, 被重置为某一个了。。
# property 可以捕获 ref_task_name 变量。
    def func(cls):
        cls._ref_tasks = tasks
        for ref_task_name in cls._ref_tasks:
            setattr(cls, ref_task_name, property(wrap_cls(ref_task_name)))

            # TODO 根据当前日期返回。
            task_name = "%s_%s" % (ref_task_name, "task")
            setattr(cls, task_name, property(wrap_instance(ref_task_name, task_name)))

            # clear ref task info when pickle.dump
            setattr(cls, "__getstate__", __getstate__)
        return cls
    return func
luigi.ref_tasks = ref_tasks



def check_date_range(): # 装饰器
    """
    从数据库导数据时，必须注意时间范围内的所有数据是否都齐全了。如果未齐全，
    即在当前时间范围里导的话，那么就会缺失数据了，相当于提前导了。

    比如在周六就把这周的关联数据导出来，那么周日的数据就没包含在里面。应该在下周一后才开始导。
    """
    def decorator(orig_run):
        def new_run(self):
            # 说明时间未到，然后就直接退出
            if self.date_value_by_type_in_begin <= ArrowParameter.now() < self.date_value_by_type_in_end:
                return False
            return orig_run(self)
        return new_run

    def func(cls):
        cls.run = decorator(cls.run)
        return cls
    return func
# TODO support Hadoop
luigi.check_date_range = check_date_range


def check_runtime_range(**opts_1): # 装饰器
    """
    Support hour/weekday indexed range.

    Optional params:
    1. hour_num
    2. weekday_num
    3. now
    """
    def decorator(orig_run):
        def new_run(self):
            default_opts = {
                    "hour_num"      : range(1, 25),
                    "weekday_num"   : range(1, 8),
                }
            opts = dict(default_opts.items() + opts_1.items())

            now            = ArrowParameter.now()           # get current time
            hour_24        = int(now.format("H"))  # 0, 1, 2, ..., 23, 24
            day_of_week_7  = int(now.format("d"))  # 1, 2, 3, ..., 6, 7

            is_false = False
            if hour_24       not in opts['hour_num']:    is_false = True
            if day_of_week_7 not in opts['weekday_num']: is_false = True
            if is_false:
                print "[info]", now, " is not in ", opts, ", so the task exited."
                return False

            return orig_run(self)
        return new_run

    def func(cls):
        cls.run = decorator(cls.run)
        return cls
    return func
luigi.check_runtime_range = check_runtime_range



def mr_local(**opts):
    """
    Sometimes Hadoop streaming sucks, so we only use the solid HDFS, and turn MapReduce job into local mode.

    And `mr_local` is optimized by a fixed chunk write operation.
    """

    def mr_run(self):
        """ Overwrite BaseHadoopJobTask#run function. """
# TODO maybe model cache
        from etl_utils import process_notifier
        map_kv_dict = defaultdict(list)

        inputs = self.input()
        if not isinstance(inputs, list): inputs = [inputs]
        for input_hdfs_1 in inputs:
            for line2 in TargetUtils.line_read(input_hdfs_1):
                for map_key_3, map_val_3 in self.mapper(line2):
                    map_kv_dict[map_key_3].append(map_val_3)

        with self.output().open("w") as output1:
            fixed_chunk = list()
            for reduce_key_2 in process_notifier(map_kv_dict.keys()):
                reduce_vals_2 = map_kv_dict[reduce_key_2]
                for _, reduce_val_2 in self.reducer(reduce_key_2, reduce_vals_2):
                    fixed_chunk.append(reduce_val_2)

                    if len(fixed_chunk) % self.chunk_size == 0:
                        output1.write("\n".join(fixed_chunk) + "\n")
                        fixed_chunk = list()
                del map_kv_dict[reduce_key_2]
            output1.write("\n".join(fixed_chunk) + "\n")


    def wrap(cls):
        cls.run = mr_run
        cls.run_mode = "mr_local"

        opts["chunk_size"] = opts.get("chunk_size", 100)
        for k1, v1 in opts.iteritems():
            setattr(cls, k1, v1)

        return cls
    return wrap
luigi.mr_local = mr_local # bind it.





def plug_packages(*package_names):
    """
    Let luigi know which packages should be attached, and can send to YARN, etc.

    Package format can be any valid Python package name, such as "project_B" or "project_C==0.0.2", etc.

    Usage: use `active_packages` decorator to notice luigi that these packages should include.
    """
    #if len(manager.luiti_config.attached_package_names) > 1:
    #    return False # default is luiti. and can plug only once.

    for p1 in package_names:
        # load all packages's depended pacakges.
        manager.luiti_config.attached_package_names.add(p1)
# TODO why should do `luigi.hadoop.attach` in `active_packages`
luigi.plug_packages = plug_packages



orig_create_packages_archive = luigi.hadoop.create_packages_archive
def create_packages_archive_with_support_egg(packages, filename):
    """
    Fix original luigi's `create_packages_archive` cannt attach egg packages
    (zip file type) to tarfile, Cause it's coping file mechanism by absolute
    path.
    """
    # 1. original create tar file
    orig_create_packages_archive(packages, filename)

    # 2. append python egg packages that 1. not covered
    import tarfile
    tar = tarfile.open(filename, "a")

    logger = luigi.hadoop.logger
    fake_exists_path = "/" # root is awlays exists
    def get_parent_zip_file_within_absolute_path(path1):
        path2      = path1[:]
        is_success = False
        while path2 != fake_exists_path:
            path2 = os.path.dirname(path2)
            if os.path.isfile(path2):
                is_success = True
                break
        return is_success, path2

    def add(src, dst):
        logger.debug('adding to tar: %s -> %s', src, dst)
        tar.add(src, dst)

    import zipfile
    import tempfile
    for package1 in packages:
        path2 = (getattr(package1, "__path__", []) + [fake_exists_path])[0]
        if os.path.exists(path2):     continue # so luigi can import it.
        if not path2.startswith("/"): continue # we only care about libraries.

        is_success, zipfilename3 = get_parent_zip_file_within_absolute_path(path2)
        if is_success:
            tmp_dir3 = tempfile.mkdtemp()
            zipfile.ZipFile(zipfilename3).extractall(tmp_dir3)

            for root4, dirs4, files4 in os.walk(tmp_dir3):
                curr_dir5 = os.path.basename(root4)
                for file5 in files4:
                    if file5.endswith(".pyc"): continue
                    add(os.path.join(root4, file5), os.path.join(root4.replace(tmp_dir3, "").lstrip("/"), file5))

    client_cfg = os.path.join(os.getcwd(), "client.cfg")
    if os.path.exists(client_cfg):
        tar.add(client_cfg, "client.cfg")
    tar.close()
luigi.hadoop.create_packages_archive = create_packages_archive_with_support_egg




luigi.ensure_active_packages = lambda : manager.active_packages # make a wrap
luigi.luiti_config = manager.luiti_config
manager.luiti_config.linked_luigi = luigi
